package operation

import (
	"context"
	"fmt"
	"time"

	"git.perx.ru/perxis/perxis-go/id"
	"git.perx.ru/perxis/perxis-go/pkg/errors"
	grpcerr "git.perx.ru/perxis/perxis-go/pkg/errors/grpc"
	"git.perx.ru/perxis/perxis-go/proto/common"
	"google.golang.org/grpc"
	"google.golang.org/protobuf/proto"
	"google.golang.org/protobuf/types/known/anypb"
	"google.golang.org/protobuf/types/known/timestamppb"
)

// DefaultPollInterval интервал опроса операции по умолчанию
const DefaultPollInterval = time.Second

type Client = common.OperationServiceClient
type Proto = common.Operation

func NewClient(conn *grpc.ClientConn) Client {
	return common.NewOperationServiceClient(conn)
}

// Operation обертка над `common.Operation` для работы с продолжительными операциями.
// Позволяет получить результат операции, проверить статус, отменить операцию
type Operation struct {
	proto  *Proto
	client Client

	retention  time.Duration
	cancelFunc func()
}

// New создает новую операцию
func New(proto *Proto) *Operation {
	o := &Operation{
		proto: proto,
	}
	o.proto.Id = id.GenerateNewID()
	o.proto.CreatedAt = timestamppb.New(time.Now())
	return o
}

// Wrap оборачивает `common.Operation` в Operation для работы c сервисом поддерживающим продолжительных операций
func Wrap(client Client, proto *Proto) *Operation {
	if proto == nil || proto.GetId() == "" {
		return nil
	}

	return &Operation{
		proto:  proto,
		client: client,
	}
}

// String возвращает строковое представление операции
func (o *Operation) String() string {
	return fmt.Sprintf("Operation #%s (%s, %s) Done:%t", o.Id(), o.CreatedAt(), o.CreatedBy(), o.IsDone())
}

// Id возвращает идентификатор операции
func (o *Operation) Id() string {
	return o.proto.GetId()
}

// Proto возвращает `common.Operation`
func (o *Operation) Proto() *Proto {
	return o.proto
}

// Description возвращает описание операции
func (o *Operation) Description() string {
	return o.proto.GetDescription()
}

// CreatedAt возвращает время создания операции
func (o *Operation) CreatedAt() time.Time {
	c := o.proto.GetCreatedAt()
	if c != nil {
		return c.AsTime()
	}
	return time.Time{}
}

// IsDone возвращает true если операция завершена
func (o *Operation) IsDone() bool {
	return o.proto.GetDone()
}

// Response возвращает результат операции
func (o *Operation) Response() (proto.Message, error) {
	r := o.proto.GetResponse()
	if r == nil {
		return nil, nil
	}
	return r.UnmarshalNew()
}

// CreatedBy возвращает имя пользователя создавшего операцию
func (o *Operation) CreatedBy() string {
	return o.proto.GetCreatedBy()
}

// SetCreatedBy устанавливает описание операции
func (o *Operation) SetCreatedBy(createdBy string) {
	o.proto.CreatedBy = createdBy
}

// SetCreatedAt устанавливает время создания операции
func (o *Operation) SetCreatedAt(t time.Time) {
	o.proto.CreatedAt = timestamppb.New(t)
}

// IsExpired возвращает true если операция просрочена
func (o *Operation) IsExpired() bool {
	createdAt := o.CreatedAt()
	if o.retention != 0 && !createdAt.IsZero() && time.Since(createdAt) > o.retention {
		return true
	}
	return false
}

func (o *Operation) SetModifiedAt(t time.Time) {
	o.proto.ModifiedAt = timestamppb.New(t)
}

// SetResponse устанавливает результат операции
func (o *Operation) SetResponse(m proto.Message) error {
	resp, err := anypb.New(m)
	if err != nil {
		return err
	}

	o.proto.Result = &common.Operation_Response{Response: resp}
	o.proto.Done = true
	return nil
}

// SetCancelFunc устанавливает функцию отмены операции
func (o *Operation) SetCancelFunc(cancelFunc func()) {
	o.cancelFunc = cancelFunc
}

// SetError устанавливает ошибку операции
func (o *Operation) SetError(err error) {
	o.proto.Result = &common.Operation_Error{Error: grpcerr.ProtoFromError(err)}
}

// Error возвращает ошибку операции
func (o *Operation) Error() error {
	return grpcerr.ErrorFromProto(nil, o.proto.GetError())
}

func (o *Operation) SetDone(done bool) {
	o.proto.Done = done
}

// SetRetention устанавливает время хранения операции
func (o *Operation) SetRetention(retention time.Duration) {
	o.retention = retention
}

// Cancel отменяет операцию
func (o *Operation) Cancel() error {
	req := &common.CancelOperationRequest{OperationId: o.Id()}
	op, err := o.client.Cancel(context.Background(), req)
	if err != nil {
		return err
	}
	o.proto = op
	return nil
}

// Poll запрашивает состояние операции
func (o *Operation) Poll(ctx context.Context, opts ...grpc.CallOption) error {
	req := &common.GetOperationRequest{OperationId: o.Id()}
	op, err := o.client.Get(ctx, req, opts...)
	if err != nil {
		return err
	}
	o.proto = op
	return nil
}

// Wait ожидает завершения операции
func (o *Operation) Wait(ctx context.Context, opts ...grpc.CallOption) error {
	return o.WaitInterval(ctx, DefaultPollInterval, opts...)
}

// WaitInterval ожидает завершения операции с указанным интервалом опроса
func (o *Operation) WaitInterval(ctx context.Context, pollInterval time.Duration, opts ...grpc.CallOption) error {
	for !o.IsDone() {
		if err := o.Poll(ctx, opts...); err != nil {
			return errors.Wrap(err, "operation poll failed")
		}

		if o.IsDone() {
			break
		}

		interval := pollInterval
		timer := time.NewTimer(interval)
		select {
		case <-timer.C:
		case <-ctx.Done():
			timer.Stop()
			return errors.Wrap(ctx.Err(), "operation wait timeout")
		}
	}

	return errors.Wrap(o.Error(), "operation failed")
}
