package setup

import (
	"io/fs"

	"git.perx.ru/perxis/perxis-go"
)

type Clonable[T any] interface {
	GetID() string
	Clone() T
}

type Conf[C any, T Clonable[T]] interface {
	Init(e *Entity[C, T])
}

type ValueFunc[T Clonable[T]] func(s *Setup, e T)
type FilterFunc[T Clonable[T]] func(t T) bool
type UpdateFunc[T Clonable[T]] func(s *Setup, exist, new T) (T, bool)
type DeleteFunc[T Clonable[T]] func(s *Setup, e T) bool
type EntityOption[C any, T Clonable[T]] func(c *Entity[C, T])

type Entity[C any, T Clonable[T]] struct {
	value      T
	ValueFunc  []ValueFunc[T]
	UpdateFunc UpdateFunc[T]
	DeleteFunc DeleteFunc[T]
	Conf       C
	Err        error
}

func NewEntity[C any, T Clonable[T]](val T, opt ...EntityOption[C, T]) *Entity[C, T] {
	e := &Entity[C, T]{value: val}

	var conf any = e.Conf
	if vv, ok := conf.(Conf[C, T]); ok {
		vv.Init(e)
	}

	for _, o := range opt {
		o(e)
	}
	return e
}

// Clone возвращает копию сущности
func (e *Entity[C, T]) Clone() *Entity[C, T] {
	return &Entity[C, T]{
		value:      e.value.Clone(),
		ValueFunc:  e.ValueFunc,
		UpdateFunc: e.UpdateFunc,
		DeleteFunc: e.DeleteFunc,
		Conf:       e.Conf,
		Err:        e.Err,
	}
}

// Value возвращает значение сущности
func (e *Entity[C, T]) Value(s *Setup) T {
	ent := e.value.Clone() // TODO: Мы уже используем копию всей конфигурации, нужно ли это?
	for _, fn := range e.ValueFunc {
		fn(s, ent)
	}
	return ent
}

type EntityList[C any, T Clonable[T]] []*Entity[C, T]

// Clone возвращает копию списка сущностей
func (l *EntityList[C, T]) Clone() EntityList[C, T] {
	var res EntityList[C, T]
	for _, e := range *l {
		res = append(res, e.Clone())
	}
	return res
}

// Add добавляет сущности в список EntityList
func (l *EntityList[C, T]) Add(t T, opt ...EntityOption[C, T]) {
	e := NewEntity(t.Clone(), opt...)
	*l = append(*l, e)
}

// AddMany добавляет несколько сущностей в список EntityList
func (l *EntityList[C, T]) AddMany(t []T, opt ...EntityOption[C, T]) {
	for _, item := range t {
		l.Add(item, opt...)
	}
}

// Filter возвращает список сущностей, удовлетворяющих фильтру
func (l *EntityList[C, T]) Filter(filter FilterFunc[T]) *EntityList[C, T] {
	res := make(EntityList[C, T], 0)
	for _, e := range *l {
		if filter(e.value) {
			res = append(res, e)
		}
	}
	return &res
}

// Get возвращает конфигурацию по ID
func (l *EntityList[C, T]) Get(id string) *Entity[C, T] {
	for _, e := range *l {
		if e.value.GetID() == id {
			return e
		}
	}
	return nil
}

// GetIDs возвращает список ID сущностей
func (l *EntityList[C, T]) GetIDs() []string {
	var res []string
	for _, e := range *l {
		res = append(res, e.value.GetID())
	}
	return res
}

// Load загружает сущности в список EntityList из указанной файловой системы
func (l *EntityList[C, T]) Load(fsys fs.FS, dir string, opt ...EntityOption[C, T]) error {
	assets := perxis.NewAssets[T](fsys)
	items, err := assets.FromDir(dir)
	if err != nil {
		return err
	}
	for _, item := range items {
		l.Add(item, opt...)
	}
	return nil
}

// MustLoad загружает сущности в список EntityList из указанной файловой системы
func (l *EntityList[C, T]) MustLoad(fsys fs.FS, dir string, opt ...EntityOption[C, T]) {
	if err := l.Load(fsys, dir, opt...); err != nil {
		panic(err)
	}
}

// WithOptions добавляет опции к сущности в списке EntityList согласно фильтру
func (l *EntityList[C, T]) WithOptions(filter FilterFunc[T], opt ...EntityOption[C, T]) {
	for _, e := range *l {
		if filter(e.value) {
			for _, o := range opt {
				o(e)
			}
		}
	}
}

// Overwrite перезаписывает сущность при обновлении
func Overwrite[C any, T Clonable[T]]() EntityOption[C, T] {
	return func(c *Entity[C, T]) {
		c.UpdateFunc = func(s *Setup, old, new T) (T, bool) { return new, true }
	}
}

// Keep сохраняет сущность при обновлении
func Keep[C any, T Clonable[T]]() EntityOption[C, T] {
	return func(c *Entity[C, T]) {
		c.UpdateFunc = func(s *Setup, old, new T) (T, bool) { return old, false }
	}
}

// Delete удаляет сущность
func Delete[C any, T Clonable[T]]() EntityOption[C, T] {
	return func(c *Entity[C, T]) {
		c.DeleteFunc = func(s *Setup, e T) bool { return true }
	}
}

// DeleteIfRemoveFlag удаляет сущность если флаг Setup Remove установлен
func DeleteIfRemoveFlag[C any, T Clonable[T]]() EntityOption[C, T] {
	return func(c *Entity[C, T]) {
		c.DeleteFunc = func(s *Setup, e T) bool { return s.IsRemove() }
	}
}
